#include <linux/clk.h>
#include <linux/init.h>
#include <linux/gpio.h>
#include <linux/platform_device.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <asm/uaccess.h>

#define AT91_BASE_ADC (0xf804c000) /* ADC base address */

#define AT91_ADC_CR   (0x00) /* ADC control register */
#define AT91_ADC_SWRST (1 << 0) /* Software reset */
#define AT91_ADC_START (1 << 1) /* Start conversion */

#define AT91_ADC_MR   (0x04) /* ADC mode register */
#define AT91_ADC_SHTIM   (0xf << 24) /* Simple & hold time */
#define AT91_ADC_STARTUP (0X0f << 16) /* Start up time */
#define AT91_ADC_PRESCAL (0xff << 8) /* Prescal rate selection */
#define AT91_ADC_SLEEP   (1 << 5) /* Sleep mode */

#define AT91_ADC_CHER (0x10) /* ADC channel enable register */
#define AT91_ADC_CH(n) (1 << n) /* Select channel */

#define AT91_ADC_ISR (0x30) /* ADC interrupt status register */
#define AT91_ADC_EOC(n) (1 << n) /* End of conversion on chanel n */

#define AT91_ADC_CDR(n) (0x50 + ((n) * 4)) /* ADC channel data register n */

#define AT91_ADC_DATA      (0x3ff)
#define AT91_ADC_DEFCONFIG AT91_ADC_SHTIM | \
	                       AT91_ADC_STARTUP | \
	                       AT91_ADC_PRESCAL | \
	                       AT91_ADC_SLEEP

#define DRV_CLASS    "at91-adc"
#define CHANNEL_NUM  12

#define ADC_IOC_MAGIC      'j'
#define ADC_REQUEST        _IOWR(ADC_IOC_MAGIC, 1, int)
#define ADC_READ           _IOR(ADC_IOC_MAGIC, 2, int)
#define ADC_FREE           _IOWR(ADC_IOC_MAGIC, 3, int)

#define at91_adc_read(reg)        ioread32(at91_adc_base + (reg))
#define at91_adc_write(reg, val)  iowrite32((val), at91_adc_base + (reg))

static const unsigned int PIN_CHAN[CHANNEL_NUM] = {
	AT91_PIN_PB11, AT91_PIN_PB12, AT91_PIN_PB13,
	AT91_PIN_PB14, AT91_PIN_PB15, AT91_PIN_PB16
};

static void at91_adc_device_release(struct device *dev)
{
}

static struct platform_device at91_adc_device = {
	.name        = "at91_adc",
	.id          = -1,
	.dev.release = at91_adc_device_release,
};

struct clk             *at91_adc_clk;
void __iomem           *at91_adc_base;
void __iomem           *at91_piob_base;
static struct cdev     *at91_adc_cdev  = NULL;
static dev_t           at91_adc_devno  = 0;
static struct class    *at91_adc_class = NULL;
static u16             mux_mask        = 0;

static int at91_adc_config(int requested_config)
{
	at91_adc_write(AT91_ADC_CR, AT91_ADC_SWRST);   // Reset ADC
	at91_adc_write(AT91_ADC_MR, requested_config); // Mode setup

	return ((requested_config == at91_adc_read(AT91_ADC_MR)) ?
			0 :
		    -EINVAL);
}

static int mux_chan(int chan, int opration)
{
	int pin_chan;
	int status;
	char gpio_label[7] = { 0 };

	if ((chan < 0) || (chan > 12)) {
		return -EINVAL;
	}

	pin_chan = PIN_CHAN[chan];

	if (opration) {
		sprintf(gpio_label, "chan%d", chan);
		if ((status = gpio_request(pin_chan, gpio_label)) < 0) {
			return status;
		}
		if ((status = gpio_direction_input(pin_chan)) < 0) {
			return status;
		}
		mux_mask |= (1 << chan);
	} else {
		gpio_free(pin_chan);
		mux_mask &= ~(1 << chan);
	}

	return 0;
}

static int at91_adc_read_chan(int chan)
{
	if ((chan < 0) || (chan > CHANNEL_NUM)) {
		printk("Unexpect channel number\n");
		return -EINVAL;
	}

	iowrite32(1 << chan, at91_piob_base + 0x60);

	at91_adc_write(AT91_ADC_CHER, AT91_ADC_CH(chan));
	at91_adc_write(AT91_ADC_CR, AT91_ADC_START);

	while (!(at91_adc_read(AT91_ADC_ISR) & AT91_ADC_EOC(chan))) {
		cpu_relax();
	}

	return at91_adc_read(AT91_ADC_CDR(chan)) & AT91_ADC_DATA;
}

static int at91_adc_chanX_show(struct device *dev,
							   struct device_attribute *attr,
							   char *buf)
{
	int value;
	int chan = -1;

	if ((strlen(attr->attr.name) == 5) &&
		(strncmp(attr->attr.name, "chan", 4) == 0)) {
		chan = attr->attr.name[4] - '0';
	}
	if ((strlen(attr->attr.name) == 6) &&
		(strncmp(attr->attr.name, "chan", 4) == 0)) {
		chan = (attr->attr.name[4] - '0') * 10 + (attr->attr.name[5] - '0');
	}

	value = at91_adc_read_chan(chan);
	if (value < 0) {
		return -EIO;
	}

	return sprintf(buf, "%d\n", value);
}

static DEVICE_ATTR(chan0, 0444, at91_adc_chanX_show, NULL);
static DEVICE_ATTR(chan1, 0444, at91_adc_chanX_show, NULL);

static const struct attribute *at91_adc_dev_attrs[] = {
	&dev_attr_chan0.attr,
	&dev_attr_chan1.attr,
	NULL,
};

static const struct attribute_group at91_adc_dev_attr_group = {
	.attrs = (struct attribute **)at91_adc_dev_attrs,
};

#ifdef HAVE_UNLOCKED_IOCTL
static long at91_adc_unlocked_ioctl(struct file   *file,
									unsigned int  cmd,
									unsigned long arg)
{
#else
static int at91_adc_ioctl(struct inode  *inode,
						  struct file   *file,
						  unsigned int  cmd,
						  unsigned int  arg)
{
#endif
	long retval = 0;
	switch (cmd) {
	case ADC_REQUEST:
		retval = mux_chan(arg, 1);
		break;
	case ADC_READ:
		retval = at91_adc_read_chan(arg);
		break;
	case ADC_FREE:
		retval = mux_chan(arg, 0);
		break;
	default:
		retval = -EINVAL;
		break;
	}

	return retval;
}

struct file_operations at91_adc_fops = {
	.owner = THIS_MODULE,
#ifdef HAVE_UNLOCKED_IOCTL
	.unlocked_ioctl = at91_adc_unlocked_ioctl,
#else
	.ioctl = at91_adc_ioctl,
#endif
};

static void at91_adc_cdev_teardown(void)
{
	unsigned int ind;

	if (at91_adc_class) {
		device_destroy(at91_adc_class, at91_adc_devno);
		class_destroy(at91_adc_class);
	}

	if (at91_adc_devno) {
		unregister_chrdev_region(at91_adc_devno, 1);
		if (at91_adc_cdev) {
			cdev_del(at91_adc_cdev);
		}
	}

	for (ind = 0; ind < CHANNEL_NUM;++ind) {
		mux_chan(ind, 0);
	}

	at91_adc_devno = 0;
	at91_adc_cdev  = NULL;
	at91_adc_class = NULL;

	return;
}

static int at91_adc_cdev_setup(void)
{
	int status;

	status = alloc_chrdev_region(&at91_adc_devno, 0, 1, at91_adc_device.name);
	if (status) {
		goto err;
	}

	/* create a new char device */
	at91_adc_cdev = cdev_alloc();
	if (at91_adc_cdev == NULL) {
		status = -ENOMEM;
		goto err;
	}
	at91_adc_cdev->owner = THIS_MODULE;
	at91_adc_cdev->ops   = &at91_adc_fops;
	status = cdev_add(at91_adc_cdev, at91_adc_devno, 1);
	if (status) {
		goto err;
	}

	/* register the class */
	at91_adc_class = class_create(THIS_MODULE, DRV_CLASS);
	if (IS_ERR(at91_adc_class)) {
		status = -EFAULT;
		goto err;
	}
	if (IS_ERR(at91_adc_class)) {
		status = -EFAULT;
		goto err;
	}
	device_create(at91_adc_class, NULL, at91_adc_devno,
				  NULL, at91_adc_device.name);
	printk(KERN_INFO "Major: %u; minor: %u\n",
		   MAJOR(at91_adc_devno),
		   MINOR(at91_adc_devno));

	return 0;

 err:
	at91_adc_cdev_teardown();
	return status;
}

static int __init at91_adc_init(void)
{
	int status;

	at91_adc_clk = clk_get(NULL, "adc_clk");
	clk_enable(at91_adc_clk);

	at91_adc_base = ioremap(AT91_BASE_ADC, SZ_256);
	if (!at91_adc_base) {
		status = -ENODEV;
		goto fail_no_iomem_adc;
	}

	at91_piob_base = ioremap(AT91_BASE_SYS + AT91_PIOB, SZ_512);
	if (!at91_piob_base) {
		status = -ENODEV;
		goto fail_no_iomem_piob;
	}

	status = platform_device_register(&at91_adc_device);
	if (status) {
		goto fail_no_dev;
	}

	status = at91_adc_config(AT91_ADC_DEFCONFIG);
	if (status) {
		goto fail_no_config;
	}

	status = sysfs_create_group(&(at91_adc_device.dev.kobj),
							  &at91_adc_dev_attr_group);
	if (status) {
		goto fail_no_sysfs;
	}

	status = at91_adc_cdev_setup();
	if (status) {
		goto fail_no_cdev;
	}

	printk(KERN_INFO "Register at91 adc device was successful\n");

	return 0;
	
 fail_no_cdev:
 fail_no_sysfs:
 fail_no_config:
	platform_device_unregister(&at91_adc_device);
 fail_no_dev:
	iounmap(at91_piob_base);
 fail_no_iomem_piob:
	iounmap(at91_adc_base);
 fail_no_iomem_adc:
	clk_disable(at91_adc_clk);
	clk_put(at91_adc_clk);

	return status;
}

static void __exit at91_adc_exit(void)
{
	at91_adc_cdev_teardown();
	platform_device_unregister(&at91_adc_device);
	iounmap(at91_adc_base);
	iounmap(at91_piob_base);

	clk_disable(at91_adc_clk);
	clk_put(at91_adc_clk);
}

module_init(at91_adc_init);
module_exit(at91_adc_exit);

MODULE_LICENSE("GPL");
 
	 
